/*
Bullet Continuous Collision Detection and Physics Library
Copyright (c) 2015 Google Inc. http://bulletphysics.org

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from the use of this software.
Permission is granted to anyone to use this software for any purpose, 
including commercial applications, and to alter it and redistribute it freely, 
subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.
*/


#include "Problem1.h"

#include "btBulletDynamicsCommon.h"
#define ARRAY_SIZE_Y 5
#define ARRAY_SIZE_X 5
#define ARRAY_SIZE_Z 5

#include "LinearMath/btVector3.h"
#include "LinearMath/btAlignedObjectArray.h"

#include "BulletDynamics/ConstraintSolver/btSliderConstraint.h"
#include "BulletDynamics/ConstraintSolver/btGeneric6DofConstraint.h"

#include "../CommonInterfaces/CommonRigidBodyBase.h"
#include "../RenderingExamples/TimeSeriesCanvas.h"


class FrictionConstraint : public btGeneric6DofConstraint{
public:
	FrictionConstraint( btRigidBody &rbA, btRigidBody &rbB, const btTransform &frameInA, const btTransform &frameInB ) :
	btGeneric6DofConstraint( rbA, rbB, frameInA, frameInB, true ){
	}
	
	virtual void getInfo2( btConstraintInfo2 *info ){
		// set up motor to apply friction relative to constraint movement
		if( m_linearLimits.m_enableMotor[ 1 ] ){
			// platform weight: 1kg
			// gravity: 10
			// force pulling down: F=a*m*dt
			// distance travelled: s=a*dt*dt
			// required coefficient: k=F/s = a*m*dt/(a*dt*dt) = m/t = m*fps
			// using 60 fps yields 60[N/s] as the coefficient where the gravity is barely compensated
			// putting a X kg box in it yields the force F=a*(m+X)*dt thus k=(m+X)*fps
			// for a 1kg box this means 120[N/s] is able to barely support the box without moving.
			// for this example 200[N/s] is used. this means the platform should stay at rest and
			// even putting a 1kg box on the platform does not make it moving. a 2kg box makes
			// the platform moving.
			// 
			// problem: bullet keeps the platform moving down without a box on it although it should stay at rest
			//          even if the coefficient is set rediculously high like 200000 this is not working
			const btScalar coefficient = ( btScalar )200;
			
			const btScalar timeStep = BT_ONE / info->fps;
			const btVector3 velocity( m_rbA.getLinearVelocity() - m_rbB.getLinearVelocity() );
			const btVector3 externalForceInvMass( m_rbA.getTotalForce() * m_rbA.getInvMass() - m_rbB.getTotalForce() * m_rbB.getInvMass() );
			
			const btVector3 axis( m_calculatedTransformA.getBasis().getColumn( 1 ) );
			const btScalar friction = btFabs( velocity.dot( axis ) + externalForceInvMass.dot( axis ) * timeStep ) * timeStep * coefficient;
			
			m_linearLimits.m_targetVelocity[ 1 ] = BT_ZERO;
			m_linearLimits.m_maxMotorForce[ 1 ] = friction;
		}
		
		// original behavior
		btGeneric6DofConstraint::getInfo2( info );
	}
};

struct Problem1InternalData{
	btRigidBody *bodyFixed;
	btRigidBody *bodyPlatform;
	btRigidBody *bodyBox;
	btGeneric6DofConstraint *constraint;
	btDefaultMotionState *motionState;
	btJointFeedback *jointFeedback;
	TimeSeriesCanvas *canvasTimeSeries;
	float dt;
	unsigned int frameID;
	
	Problem1InternalData() : dt( 1.0 / 60.0 ), frameID( 0 ){ }
};

struct Problem1 : public CommonRigidBodyBase{
	struct Problem1InternalData* data;
	
	Problem1( struct GUIHelperInterface *helper );
	virtual ~Problem1();
	virtual void initPhysics();
	virtual void renderScene();
	virtual void stepSimulation( float deltaTime );
	void resetCamera();
};

Problem1::Problem1( struct GUIHelperInterface *helper ) : CommonRigidBodyBase( helper ){
	data = new Problem1InternalData;
	data->canvasTimeSeries = new TimeSeriesCanvas(helper->get2dCanvasInterface(), 256, 256, "Constraint Status");
	data->canvasTimeSeries ->setupTimeSeries( 20, 100, 0 );
	data->canvasTimeSeries->addDataSource( "Joint Force [N]", 255, 0, 0 );
	data->canvasTimeSeries->addDataSource( "Velocity [m/s]", 0, 0, 255 );
	data->canvasTimeSeries->addDataSource( "Position [m]", 0, 255, 0 );
}

Problem1::~Problem1(){
	exitPhysics();
	
	delete data->jointFeedback;
	data->jointFeedback = 0;
	delete data->canvasTimeSeries;
	data->canvasTimeSeries = 0;
	
	delete data;
}

void Problem1::initPhysics(){
	m_guiHelper->setUpAxis(1);
	
	m_collisionConfiguration = new btDefaultCollisionConfiguration();
	m_dispatcher = new btCollisionDispatcher( m_collisionConfiguration );
	btVector3 worldAabbMin( -10000.0, -10000.0, -10000.0 );
	btVector3 worldAabbMax( 10000.0, 10000.0, 10000.0 );
	m_broadphase = new btAxisSweep3( worldAabbMin, worldAabbMax );
	
	m_solver = new btSequentialImpulseConstraintSolver;
	//m_solver = new btNNCGConstraintSolver;
	
	m_dynamicsWorld = new btDiscreteDynamicsWorld( m_dispatcher, m_broadphase, m_solver, m_collisionConfiguration );
	m_dynamicsWorld->getDispatchInfo().m_useContinuous = true;
	m_guiHelper->createPhysicsDebugDrawer( m_dynamicsWorld );
	
	if( m_dynamicsWorld->getDebugDrawer() ){
		m_dynamicsWorld->getDebugDrawer()->setDebugMode( btIDebugDraw::DBG_DrawWireframe + btIDebugDraw::DBG_DrawContactPoints );
	}
	
	m_dynamicsWorld->getSolverInfo().m_numIterations = 20; //100;
	
	// ground shape
	{
		btCollisionShape * const shape = new btBoxShape( btVector3( 200.0, 5.0, 200.0 ) );
		btTransform transform;
		transform.setIdentity();
		transform.setOrigin( btVector3( 0.0, -5.0, 0.0 ) );
		btCollisionObject * const body = new btCollisionObject();
		body->setCollisionShape( shape );
		body->setWorldTransform( transform );
		m_dynamicsWorld->addCollisionObject( body );
	}
	
	// static rigid body to attach platform to
	{
		btCollisionShape * const shape = new btBoxShape( btVector3( 0.1, 0.1, 0.1 ) );
		btTransform transform;
		transform.setIdentity();
		data->bodyFixed = new btRigidBody( 0.0, NULL, shape, btVector3() );
		data->bodyFixed->setCollisionShape( shape );
		data->bodyFixed->setWorldTransform( transform );
		m_dynamicsWorld->addCollisionObject( data->bodyFixed );
	}
	
	{
		// create platform
		btCollisionShape * const shapeChild = new btBoxShape( btVector3( 0.85, 0.04998, 0.85 ) );
		btCompoundShape * const shape = new btCompoundShape;
		btTransform transformShapeChild;
		transformShapeChild.setIdentity();
		transformShapeChild.setOrigin( btVector3( 0.0, 0.01, 0.0 ) );
		shape->addChildShape( transformShapeChild, shapeChild );
		
		btTransform platformTransform;
		platformTransform.setIdentity();
		platformTransform.setOrigin( btVector3( 0.0, 2.05, 0.0 ) );
		
		btScalar mass = 1.0;
		btVector3 localInertia( 0.0, 0.0, 0.0 );
		shape->calculateLocalInertia( mass, localInertia );
		
		data->motionState = new btDefaultMotionState( platformTransform );
		data->bodyPlatform = new btRigidBody( mass, data->motionState, shape, localInertia );
		data->bodyPlatform->setActivationState( DISABLE_DEACTIVATION );
		m_dynamicsWorld->addRigidBody( data->bodyPlatform );
		
		// platform constraint
		btTransform localA;
		localA.setIdentity();
		localA.setOrigin( btVector3( 0.0, -0.05, 0.9 ) );
		
		btTransform localB;
		localB.setIdentity();
		localB.setOrigin( btVector3( 0.0, 2.0, 0.9 ) );
		
		const btScalar cfm = 0.0; // bullet default 0.0
		const btScalar stopCfm = 0.0; // bullet default 0.0
		const btScalar stopErp = 0.2; // bullet default 0.2
		const btScalar damping = 1.0; // bullet default 1.0
		const btScalar limitSoftness = 0.7; // bullet default 0.7
		const btScalar restitution = 0.5; // bullet default 0.5
		const btScalar bounce = 0.0; // bullet default 0.0
		
		int i;
		FrictionConstraint * const constraint =  new FrictionConstraint(
			*data->bodyPlatform, *data->bodyFixed, localA, localB );
		for( i=0; i<6; i++ ){
			constraint->setLimit( i, 0.0, 0.0 );
			constraint->setParam( BT_CONSTRAINT_CFM, cfm, i );
			constraint->setParam( BT_CONSTRAINT_STOP_CFM, stopCfm, i );
			constraint->setParam( BT_CONSTRAINT_STOP_ERP, stopErp, i );
		}
		constraint->getTranslationalLimitMotor()->m_damping = damping;
		constraint->getTranslationalLimitMotor()->m_limitSoftness = limitSoftness;
		constraint->getTranslationalLimitMotor()->m_restitution = restitution;
		for( i=3; i<6; i++ ){
			constraint->getRotationalLimitMotor( i )->m_damping = damping;
			constraint->getRotationalLimitMotor( i )->m_limitSoftness = limitSoftness;
			constraint->getRotationalLimitMotor( i )->m_bounce = bounce;
		}
		
		constraint->setLimit( 1, 0.0, 1.5 ); // linear y, this is the actual dof of interest
		constraint->getTranslationalLimitMotor()->m_enableMotor[ 1 ] = true;
		
		constraint->setDbgDrawSize( 2.0 );
		
		data->jointFeedback = new btJointFeedback;
		constraint->setJointFeedback( data->jointFeedback );
		
		m_dynamicsWorld->addConstraint( constraint, true );
	}
	
	if(false){
		// place a box ontop of the platform
		btCollisionShape * const shape = new btBoxShape( btVector3( 0.3, 0.3, 0.3 ) );
		
		btTransform transform;
		transform.setIdentity();
		transform.setOrigin( btVector3( -0.4, 2.5, -0.4 ) );
		
		btScalar mass = 4.0;
		btVector3 localInertia( 0.0, 0.0, 0.0 );
		shape->calculateLocalInertia( mass, localInertia );
		
		btDefaultMotionState * const motionState = new btDefaultMotionState( transform );
		data->bodyBox = new btRigidBody( mass, motionState, shape, localInertia );
		data->bodyBox->setActivationState( DISABLE_DEACTIVATION );
		m_dynamicsWorld->addRigidBody( data->bodyBox );
	}
	
	m_guiHelper->autogenerateGraphicsObjects( m_dynamicsWorld );
}

void Problem1::renderScene(){
	CommonRigidBodyBase::renderScene();
}

void Problem1::resetCamera(){
	float dist = 3;
	float pitch = -15;
	float yaw = 52;
	float targetPos[3] = { 0.0, 1.0, 0.0 };
	m_guiHelper->resetCamera(dist,yaw,pitch,targetPos[0],targetPos[1],targetPos[2]);
}

void Problem1::stepSimulation( float deltaTime ){
	m_dynamicsWorld->stepSimulation( deltaTime );
	
	btTransform transform;
	data->motionState->getWorldTransform( transform );
	
	printf( "appliedForce=%g linearVelocity=%g position=%g\n",
		data->jointFeedback->m_appliedForceBodyA.getY(),
		data->bodyPlatform->getLinearVelocity().getY(),
		transform.getOrigin().getY() );
	
	data->canvasTimeSeries->insertDataAtCurrentTime( data->jointFeedback->m_appliedForceBodyA.getY(), 0, true );
	data->canvasTimeSeries->insertDataAtCurrentTime( data->bodyPlatform->getLinearVelocity().getY(), 1, true );
	data->canvasTimeSeries->insertDataAtCurrentTime( transform.getOrigin().getY(), 2, true );
	data->canvasTimeSeries->nextTick();
}

CommonExampleInterface *Problem1CreateFunc( CommonExampleOptions &options ){
	return new Problem1( options.m_guiHelper );
}

B3_STANDALONE_EXAMPLE(Problem1CreateFunc)
